

class _Context:
    def __init__(self, path):
        self.path = path
        self.temp_path = self.path.with_suffix(".tmp")
        self.lock = self.path.lock()

    def __enter__(self):
        self.lock.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.lock.__exit__(exc_type, exc_val, exc_tb)


class WriteContext(_Context):
    """ A context manager used for every write method. """
    def __init__(self, path, overwrite=False):
        super().__init__(path)
        self.overwrite = overwrite

    def __enter__(self):
        """ :rtype: generalfile.Path """
        if not self.overwrite and self.path.exists():
            raise FileExistsError(f"Path '{self.path}' already exists and overwrite is 'False'.")
        super().__enter__()
        self.path.get_parent().create_folder()
        return self.temp_path

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.temp_path.rename(self.path.name(), overwrite=True)
        super().__exit__(exc_type, exc_val, exc_tb)


class ReadContext(_Context):
    """ A context manager used for every read method. """
    def __init__(self, path):
        super().__init__(path)

    def __enter__(self):
        super().__enter__()
        return self.path

    def __exit__(self, exc_type, exc_val, exc_tb):
        super().__exit__(exc_type, exc_val, exc_tb)


class AppendContext(_Context):
    """ A context manager used for every append method. """
    def __init__(self, path):
        super().__init__(path)

    def __enter__(self):
        super().__enter__()
        if self.path.exists():
            self.path.copy(self.temp_path)
        return self.temp_path

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.temp_path.rename(self.path.name(), overwrite=True)
        super().__exit__(exc_type, exc_val, exc_tb)


class _Extension:
    WriteContext = WriteContext
    ReadContext = ReadContext
    AppendContext = AppendContext

    def __init__(self, path):
        self.path = path






from generallibrary import deco_cache, initBases
from generalfile.optional_dependencies._extension import _Extension

import configparser
import json


class Path_Cfg:
    """ Cfg methods for Path. """
    @property
    @deco_cache()
    def cfg(self):
        """ Easily modify cfg files with a dictionary. """
        return _Cfg(self)


@initBases
class _Cfg(_Extension):
    _JSON_CAST = {
        "None": None,
        "True": True,
        "False": False,
    }

    def write(self, dict_=None, overwrite=False):
        """ Write to this path with a given dictionary. """
        config = configparser.RawConfigParser()
        config.read_dict(dictionary=dict_)
        with self.WriteContext(self.path, overwrite=overwrite) as write_path:
            return write_path.open_operation("w", lambda stream: config.write(stream))

    def _read_json_cast(self, value):
        if value in self._JSON_CAST:
            return self._JSON_CAST[value]
        try:
            return json.loads(value.replace("'", '"'))
        except json.decoder.JSONDecodeError:
            return value

    def read(self, default=...):
        """ Read from this path to get a dictionary. """
        config = configparser.RawConfigParser()
        with self.ReadContext(self.path) as read_path:

            if not read_path.exists():
                if default is Ellipsis:
                    raise FileNotFoundError
                else:
                    return default

            config.read(read_path)
        return {s: {key: self._read_json_cast(value) for key, value in config.items(s)} for s in config.sections()}

    def append(self, dict_):
        """ Update this cfg with a dictionary. """
        with self.AppendContext(self.path) as append_path:
            # append_path.cfg.write(dict_=append_path.cfg.read() | dict_, overwrite=True)
            append_path.cfg.write(dict_={**append_path.cfg.read(), **dict_}, overwrite=True)














































def setup_workdir():
    """ Set working dir for tests and clear it after it's made sure it's correct path."""
    repo_path = Path().get_parent_repo()

    if not repo_path:
        raise EnvironmentError(f"Failed setting correct woroking dir, current working dir is not inside a repo")

    path = repo_path / repo_path.name() / "test/tests"
    path.set_working_dir()

    if not Path.get_working_dir().endswith("test/tests"):
        raise EnvironmentError(f"Failed setting correct working dir, should be ..test/tests but it's {path}")

    path.delete_folder_content()

from generalfile.path import Path




from generalfile.test.test_path import PathTest
from generalfile.configfile import ConfigFile

from generallibrary import Ver
from generalfile import Path


class TestConfigFile(PathTest):
    def test_suffix(self):
        class A(ConfigFile):
            name = "foo"
        self.assertRaises(AssertionError, A, path="hi")
        self.assertRaises(AssertionError, A, path="json")
        self.assertRaises(AssertionError, A, path="foo.json.test")
        A(path="foo.json")
        A(path="foos.JSON")
        A(path="foo.cfg")
        A(path="foos.CFG")
        A(path="foo.test.cfg")

    def test_file(self):
        class A(ConfigFile):
            name = "foo"
        a = A(path="foo.json")

        self.assertEqual(False, a._path.exists())
        a.name = "bar"
        self.assertEqual(True, a._path.exists())

        b = A(path="foo.json")
        self.assertEqual("bar", b.name)

    def test_name(self):
        class A(ConfigFile):
            name = "foo"
        a = A(path="foo.json")
        self.assertEqual("foo", a.name)

        a.name = "bar"
        self.assertEqual("bar", a.name)

    def test_default_path(self):
        class A(ConfigFile):
            name = "foo"
            def __init__(self, path="foo.json"):
                pass
        a = A()

        self.assertEqual(False, a._path.exists())
        a.name = "bar"
        self.assertEqual(True, a._path.exists())

        b = A(path="foo.json")
        self.assertEqual("bar", b.name)

    def test_serializer_with_default(self):
        class A(ConfigFile):
            ver = Ver(0.2)
        a = A(path="foo.json")

        self.assertEqual(Ver(0.2), a.ver)
        a.write_config()
        a._read_config()
        self.assertEqual(Ver(0.2), a.ver)
        self.assertIsInstance(a.ver, Ver)

    def test_serializer_with_annotation(self):
        class A(ConfigFile):
            ver: Ver = Ver(0.2)
        a = A(path="foo.json")

        self.assertEqual(Ver(0.2), a.ver)
        a.write_config()
        a._read_config()
        self.assertEqual(Ver(0.2), a.ver)
        self.assertIsInstance(a.ver, Ver)

    def test_serializer_with_only_annotation(self):
        class A(ConfigFile):
            ver: Ver = None
        a = A(path="foo.json")

        self.assertIs(None, a.ver)
        a.write_config()
        a._read_config()
        self.assertIs(None, a.ver)

        a.ver = Ver("0.2.1")
        self.assertEqual(Ver("0.2.1"), a.ver)
        self.assertIsInstance(a.ver, Ver)

    def test_get_custom_serializers(self):
        class A(ConfigFile):
            ver: Ver = None
            ver2 = Ver(5)
            x = 2
            y = "foo"
            z = ["bar"]
            path = Path("foo")
            path2: Path = Path("bar")

            def __init__(self, path="foo.json"):
                pass

        self.assertEqual({
            "ver": Ver,
            "ver2": Ver,
            "path": Path,
            "path2": Path
        }, A(path="foo.json").get_custom_serializers())

        x = A().field_dict().copy()

        A().write_config()
        A()._read_config()

        self.assertEqual(x, A().field_dict())

    def test_recycle(self):
        class A(ConfigFile):
            pass

        self.assertIs(A("x.json"), A("x.json"))
        self.assertIsNot(A("x.json"), A("y.json"))

    def test_field_keys(self):
        class A(ConfigFile):
            ver: Ver = None
            ver2 = Ver(5)
            x = 2
            y = "foo"
            z = ["bar"]
            path = Path("foo")
            path2: Path = Path("bar")

        self.assertCountEqual([
            "ver",
            "ver2",
            "x",
            "y",
            "z",
            "path",
            "path2",
        ], A("foo.json").field_keys())

        a = A(path="foo.json")

        a.x = 3

        self.assertEqual({
            "ver": None,
            "ver2": Ver(5),
            "x": 2,
            "y": "foo",
            "z": ["bar"],
            "path": Path("foo"),
            "path2": Path("bar"),
        }, a.field_dict_defaults())

    def test_load_on_init(self):
        class A(ConfigFile):
            foo = "hi"

        Path("foo.json").write({"foo": "bar"})

        a = A("foo.json")
        self.assertEqual("bar", a.foo)

    def test_remove_unused(self):
        class A(ConfigFile):
            foo = "hi"

        Path("foo.json").write({
            "foo": "bar",
            "unused": True,
        })

        a = A("foo.json")
        self.assertEqual(a.foo, "bar")
        self.assertRaises(AttributeError, getattr, a, "unused")

        self.assertEqual({
            "foo": "bar",
            "unused": True,
        }, Path("foo.json").read())

        a.foo = "hi"

        # unused is removed at first write, dont force it
        self.assertEqual({
            "foo": "hi",
        }, Path("foo.json").read())

    def test_only_annotation(self):
        class A(ConfigFile):
            name: str
        a = A("foo.json")
        self.assertEqual([], a.field_keys())
        self.assertRaises(AttributeError, getattr, a, "name")

    def test_only_annotation_which_is_serializer(self):
        class A(ConfigFile):
            version: Ver
        a = A("foo.json")
        self.assertEqual([], a.field_keys())
        self.assertRaises(AttributeError, getattr, a, "ver")

    def test_dict(self):
        class A(ConfigFile):
            x = {"Foo": "bar", "5": 2}
        a = A(path="foo.json")

        self.assertEqual({"Foo": "bar", "5": 2}, a.x)
        a.write_config()
        a._read_config()
        self.assertEqual({"Foo": "bar", "5": 2}, a.x)





class TestConfigFileCFG(PathTest):
    def test_file(self):
        class A(ConfigFile):
            name = "foo"
        a = A(path="foo.cfg")

        self.assertEqual(False, a._path.exists())
        a.name = "bar"
        self.assertEqual(True, a._path.exists())

        b = A(path="foo.cfg")
        self.assertEqual("bar", b.name)

    def test_name(self):
        class A(ConfigFile):
            name = "foo"
        a = A(path="foo.cfg")
        self.assertEqual("foo", a.name)

        a.name = "bar"
        self.assertEqual("bar", a.name)

    def test_default_path(self):
        class A(ConfigFile):
            name = "foo"
            def __init__(self, path="foo.cfg"):
                pass
        a = A()

        self.assertEqual(False, a._path.exists())
        a.name = "bar"
        self.assertEqual(True, a._path.exists())

        b = A(path="foo.cfg")
        self.assertEqual("bar", b.name)

    def test_serializer_with_default(self):
        class A(ConfigFile):
            ver = Ver(0.2)
        a = A(path="foo.cfg")

        self.assertEqual(Ver(0.2), a.ver)
        a.write_config()
        a._read_config()
        self.assertEqual(Ver(0.2), a.ver)
        self.assertIsInstance(a.ver, Ver)

    def test_serializer_with_annotation(self):
        class A(ConfigFile):
            ver: Ver = Ver(0.2)
        a = A(path="foo.cfg")

        self.assertEqual(Ver(0.2), a.ver)
        a.write_config()
        a._read_config()
        self.assertEqual(Ver(0.2), a.ver)
        self.assertIsInstance(a.ver, Ver)

    def test_serializer_with_only_annotation(self):
        class A(ConfigFile):
            ver: Ver = None
        a = A(path="foo.cfg")

        self.assertIs(None, a.ver)
        a.write_config()
        a._read_config()
        self.assertIs(None, a.ver)

        a.ver = Ver("0.2.1")
        self.assertEqual(Ver("0.2.1"), a.ver)
        self.assertIsInstance(a.ver, Ver)

    def test_get_custom_serializers(self):
        class A(ConfigFile):
            ver: Ver = None
            ver2 = Ver(5)
            x = 2
            y = "foo"
            z = ["bar"]
            path = Path("foo")
            path2: Path = Path("bar")

            def __init__(self, path="foo.cfg"):
                pass

        self.assertEqual({
            "ver": Ver,
            "ver2": Ver,
            "path": Path,
            "path2": Path
        }, A(path="foo.cfg").get_custom_serializers())

        x = A().field_dict().copy()

        A().write_config()
        A()._read_config()

        self.assertEqual(x, A().field_dict())

    def test_recycle(self):
        class A(ConfigFile):
            pass

        self.assertIs(A("x.cfg"), A("x.cfg"))
        self.assertIsNot(A("x.cfg"), A("y.cfg"))

    def test_field_keys(self):
        class A(ConfigFile):
            ver: Ver = None
            ver2 = Ver(5)
            x = 2
            y = "foo"
            z = ["bar"]
            a = {"foo": "bar"}
            path = Path("foo")
            path2: Path = Path("bar")

        self.assertCountEqual([
            "ver",
            "ver2",
            "x",
            "y",
            "z",
            "a",
            "path",
            "path2",
        ], A("foo.cfg").field_keys())

        a = A(path="foo.cfg")

        a.x = 3

        self.assertEqual({
            "ver": None,
            "ver2": Ver(5),
            "x": 2,
            "y": "foo",
            "z": ["bar"],
            "a": {"foo": "bar"},
            "path": Path("foo"),
            "path2": Path("bar"),
        }, a.field_dict_defaults())

    def test_load_on_init(self):
        class A(ConfigFile):
            foo = "hi"

        Path("foo.cfg").cfg.write({A._CFG_HEADER_NAME: {"foo": "bar"}})

        a = A("foo.cfg")
        self.assertEqual(a.foo, "bar")

    def test_remove_unused(self):
        class A(ConfigFile):
            foo = "hi"

        Path("foo.cfg").cfg.write({A._CFG_HEADER_NAME: {
            "foo": "bar",
            "unused": True,
        }})

        a = A("foo.cfg")
        self.assertEqual(a.foo, "bar")
        self.assertRaises(AttributeError, getattr, a, "unused")

        self.assertEqual({A._CFG_HEADER_NAME: {
            "foo": "bar",
            "unused": True,
        }}, Path("foo.cfg").cfg.read())

        a.foo = "hi"

        # unused is removed at first write, dont force it
        self.assertEqual({A._CFG_HEADER_NAME: {
            "foo": "hi",
        }}, Path("foo.cfg").cfg.read())

    def test_only_annotation(self):
        class A(ConfigFile):
            name: str
        a = A("foo.cfg")
        self.assertEqual([], a.field_keys())
        self.assertRaises(AttributeError, getattr, a, "name")

    def test_only_annotation_which_is_serializer(self):
        class A(ConfigFile):
            version: Ver
        a = A("foo.cfg")
        self.assertEqual([], a.field_keys())
        self.assertRaises(AttributeError, getattr, a, "ver")

    def test_dict(self):
        class A(ConfigFile):
            x = {"Foo": "bar", "5": 2}
        a = A(path="foo.cfg")

        self.assertEqual({"Foo": "bar", "5": 2}, a.x)
        a.write_config()
        a._read_config()
        self.assertEqual({"Foo": "bar", "5": 2}, a.x)

    def test_list(self):
        class A(ConfigFile):
            x = ["foo", "bar"]
        a = A(path="foo.cfg")

        self.assertEqual(["foo", "bar"], a.x)
        a.write_config()
        a._read_config()
        self.assertEqual(["foo", "bar"], a.x)




















from generalfile import Path
from generalfile.test.test_path import PathTest


class TextTest(PathTest):
    paths = ("a", "b.txt", "c.md")

    def test_text_read_write(self):
        for path in self.paths:
            text = Path(path).text
            text.write("foo")
            self.assertEqual("foo", text.read())

            self.assertRaises(FileExistsError, text.write, "bar")

            text.write("bar", overwrite=True)
            self.assertEqual("bar", text.read())

    def test_text_append(self):
        for path in self.paths:
            text = Path(path).text

            text.append("foo")
            self.assertEqual("foo", text.read())

            text.append("bar")
            self.assertEqual("foobar", text.read())

            text.append("linebreak", newline=True)
            self.assertEqual("foobar\nlinebreak", text.read())

    def test_text_replace(self):
        for path in self.paths:
            text = Path(path).text

            text.write("hello this is a test")

            text.replace({"this": "that"})
            self.assertEqual("hello that is a test", text.read())

            text.replace({"that is": "what"})
            self.assertEqual("hello what a test", text.read())

            text.replace({"hello": "hi", "a test": "are tests"})
            self.assertEqual("hi what are tests", text.read())

    def test_text_replace_regex(self):
        for path in self.paths:
            text = Path(path).text

            text.write("hello this is a test")

            text.replace({".+": "foo52bar"}, regex=True)
            self.assertEqual("foo52bar", text.read())

            text.replace({".+": "foo52bar"}, regex=True)
            self.assertEqual("foo52bar", text.read())

            text.replace({"\d+": ""}, regex=True)
            self.assertEqual("foobar", text.read())

    def test_utf8(self):
        path = Path("foo")
        path.text.write("├")
        self.assertEqual("├", path.text.read())

    def test_read_empty(self):
        with self.assertRaises(FileNotFoundError):
            Path("hey").text.read()
        self.assertEqual(None, Path("hi").text.read(default=None))


































from generallibrary import deco_cache
from generalfile.errors import InvalidCharacterError

import re


class _Path_Scrub:
    @classmethod
    @deco_cache()
    def scrub(cls, str_path):
        """ :param generalfile.Path cls: """
        return cls._scrub(str_path=str_path)

    @classmethod
    def _scrub(cls, str_path):
        """ :param generalfile.Path cls: """
        str_path = "" if str_path is None else str(str_path)
        str_path = cls._replace_delimiters(str_path=str_path)


        match = re.match(fr"(\.+{re.escape(cls.path_delimiter)})", str_path)
        if match:
            dots = match.group()
            rest = str_path[len(dots):]
            index = len(dots) - 2

            rest_path = cls.Path(rest)
            working_dir = cls.Path().absolute()
            dot_path = working_dir.get_parent(depth=index, index=index, include_self=True)
            if dot_path is None:
                dot_path = working_dir.root()
            return str(dot_path / rest_path)


        str_path = cls._invalid_characters(str_path=str_path)
        str_path = cls._trim(str_path=str_path)
        str_path = cls._delimiter_suffix_if_root(str_path=str_path)


        return str_path

    @classmethod
    @deco_cache()
    def _replace_delimiters(cls, str_path):
        """ :param generalfile.Path cls: """
        str_path = str_path.replace("/", cls.path_delimiter)
        str_path = str_path.replace("\\", cls.path_delimiter)
        return str_path

    @classmethod
    @deco_cache()
    def _invalid_characters(cls, str_path):
        """ :param generalfile.Path cls: """
        # Simple invalid characters testing from Windows
        for character in '<>"|?*':
            if character in str_path:
                raise InvalidCharacterError(f"Invalid character '{character}' in '{str_path}'")

        if ":" in str_path:
            if not cls.verInfo.pathRootHasColon:
                raise InvalidCharacterError(f"Path has a colon but '{cls.verInfo.os}' doesn't use colon for path root: '{str_path}'")
            if str_path[1] != ":":
                raise InvalidCharacterError(f"Path has a colon but there's no colon at index 1: '{str_path}'")
            if len(str_path) >= 3 and str_path[2] != cls.path_delimiter:
                raise InvalidCharacterError(f"Path has a colon but index 2 is not a delimiter: '{str_path}'")
            if ":" in str_path[2:]:
                raise InvalidCharacterError(f"Path has a colon that's not at index 1: '{str_path}'")

        if str_path.endswith("."):
            raise InvalidCharacterError(f"Path cannot end with a dot ('.').")
        return str_path

    @classmethod
    @deco_cache()
    def _trim(cls, str_path):
        """ :param generalfile.Path cls: """
        if not cls.verInfo.pathRootIsDelimiter and str_path.startswith(cls.path_delimiter):
            str_path = str_path[1:]
        if str_path.endswith(cls.path_delimiter) and len(str_path) > 1:
            str_path = str_path[0:-1]
        return str_path

    @classmethod
    @deco_cache()
    def _delimiter_suffix_if_root(cls, str_path):
        """ :param generalfile.Path cls: """
        if len(str_path) == 2 and str_path[1] == ":":
            return f"{str_path}{cls.path_delimiter}"
        return str_path